#!/usr/bin/env python

import os
import sys
import warnings
import pandas as pd 
import numpy as np 
import argparse # pour gerer les options
import re # pour les regexp
from pprint import pprint


class gmapstats (object):

	def __init__(self):
		# initialize mains objects here
		self.parameters = None
		self.gmapResults = {}
		self.perfectMatches = []

		self.stats = { 'num_perfect_matches' : 0, 
			'num_perfect_multi' : 0,
			'num_ambigous_match':0}

		self.version='1.0'
		self.author='Helene Rimbert'

	def main(self):
		"""
		MAIN
		"""
		self.parseOptions()
		self.createWorkingEnv()
		self.getGmapResults()
		self.parseGmapResults()

		pprint(self.stats)

		self.outoutPerfectMatch.close()
		self.outoutPerfectMatchMulti.close()
		self.outoutAmbigousMatch.close()


	def parseOptions(self):
		"""
		Use Argparse to setup/get options from cmd line
		"""
		parser = argparse.ArgumentParser()
		parser.add_argument("-i", "--input", help="GMAP results in PSL format (-f 1 gmap option)", required=True)
		parser.add_argument("-o", "--outputdir", help="output directory", required=True)
		parser.add_argument("-c", "--qcov", help="min query coverage for ambigous matches", default=0.95)
		self.parameters=parser.parse_args()

	def createWorkingEnv(self):

		# create output dir
		try:
			os.mkdir(self.parameters.outputdir)
		except OSError:
			print (" Creation of directory  %s failed" % self.parameters.outputdir)
		else:
			print (" Successfully created the directory %s " % self.parameters.outputdir)

		# create filehandler for results
		self.outoutPerfectMatch = open(self.parameters.outputdir+"/perfectmatch.psl", "w")
		self.outoutPerfectMatchMulti = open(self.parameters.outputdir+"/perfectmatch_multi.psl", "w")
		self.outoutAmbigousMatch = open(self.parameters.outputdir+"/ambigous_match.psl", "w")


	def getGmapResults(self):
		"""
		Read gmap psl results file and save into dict
		"""

		pslFile = self.parameters.input
		if os.path.isfile(pslFile):
			with open(pslFile, 'r') as myFile:
				pslContent=myFile.readlines()

			for pslLine in pslContent:
				(matches,mismatches,repmatches,ncount,qnuminsert,qbaseinsert,tnuminsert,tbaseinsert,strand,qname,qsize,qstart,qend,tname,tsize,tstart,tend,blockcount,blocksize,qstart,tstart) = pslLine.split("\t")
				
				if qname in self.gmapResults:
					print("     + new gmap results for query %s " % str(qname))
					self.gmapResults[qname].append(pslLine)
				else:
					print(" >>>>>>>>>> new Query %s " % str(qname))
					self.gmapResults[qname] = [pslLine]


	def getInitialChrom(self, query):
		"""
		extract chromosome from query name
		"""
		queryinfo = query.split(':')
		chrom=queryinfo[2]
		return chrom

	def parseGmapResults(self):
		"""
		foreach query, parse result and check if we are having a perfect match or not
		"""

		queryList = list(self.gmapResults.keys())
		numQueries = len(queryList)
		print(" Found %i queries in GMAP results " % numQueries)

		for query in queryList:
			if query == "qname":
				continue

			print( "\n\n Evaluate query %s" % str(query))

			# get dict of gmap results
			gmapresults = self.gmapResults[query]
			nhits = len(gmapresults)
			print(" %i results found" % nhits)
			#pprint(gmapresults)

			# create dict of perfect hit(s) for current query
			self.perfectMatches = []

			# loop over all results
			for hit in gmapresults:
				(matches,mismatches,repmatches,ncount,qnuminsert,qbaseinsert,tnuminsert,tbaseinsert,strand,qname,qsize,qstart,qend,tname,tsize,tstart,tend,blockcount,blocksize,qstart,tstart) = hit.split("\t")
				
				# get percentage of query coverage
				qcoverage = (int(matches) / int(qsize)) * 100 

				# extract chromosome from query (based on query name)
				initialChrom = self.getInitialChrom(query=qname)

				# check if current hit is a perfect match
				self.checkPerfectMatch(hit=hit, qcov=qcoverage, initChrom=initialChrom)
			
			numPerfectMatches = len(self.perfectMatches)
			print(" found %i perfect matches" % numPerfectMatches)

			if numPerfectMatches == 1:
				self.stats['num_perfect_matches'] += 1
				self.outputMatches(matches=self.perfectMatches, category='perfect')

			elif numPerfectMatches >1:
				self.stats['num_perfect_multi'] += 1
				self.outputMatches(matches=self.perfectMatches, category='multi')

			else:
				self.stats['num_ambigous_match'] += 1
				self.outputMatches(matches=gmapresults, category='ambigous')

	def outputMatches(self,matches,category):

		if category == 'multi':
			for hit in matches:
				self.outoutPerfectMatchMulti.write(str(hit))

		elif category == 'perfect':
			self.outoutPerfectMatch.write(str(matches[0]))

		elif category == 'ambigous':
			for hit in matches:
				self.outoutAmbigousMatch.write(str(hit))

		else:
			print(" ERROR: problem with category of gmap hits")
			print("        Please check ")
			#pprint(matches)
			sys.exit()

	def getInitialChrom(self, query):
		"""
		extract chromosome from query name
		"""
		queryinfo = query.split(':')
		chrom=queryinfo[2]
		return chrom.lower()	

	def checkPerfectMatch(self, hit, qcov, initChrom):
		(matches,mismatches,repmatches,ncount,qnuminsert,qbaseinsert,tnuminsert,tbaseinsert,strand,qname,qsize,qstart,qend,tname,tsize,tstart,tend,blockcount,blocksize,qstart,tstart) = hit.split("\t")
		tname = tname.lower()

		# print( " -----------------------------------------")
		# print( " target: %s" % tname)
		# print( " query coverage: %s" % str(qcov))
		# print( " query initial chromosome: %s" % str(initChrom))
		# print( " target length: %s" % tsize)
		# print( " query length: %s" % qsize)
		# print( " num matches: %s" % matches)
		# print( " num mismatches: %s" % mismatches)
		# print( " num repmatches: %s" % repmatches)
		# print( " num ncount: %s" % ncount)
		# print( " num query insert: %s" % qnuminsert)
		# print( " num query base insert: %s" % qbaseinsert)
		# print( " num target insert: %s" % tnuminsert)
		# print( " num target base insert: %s" % tbaseinsert)

		if 	(int(qcov) == 100 and tname == initChrom and int(mismatches) == 0 and  int(repmatches) == 0 and  int(ncount) == 0 and  int(qnuminsert) == 0 and  int(qbaseinsert) == 0 and int(tnuminsert) == 0 and  int(tbaseinsert) == 0) :
			"""
			perfect Match found
			"""
			print("     >>>>>>>>>>>>>>>> Perfect match against target %s <<<<<<<<<<<<<<<<<<" % tname)

			self.perfectMatches.append(hit)

if __name__ == "__main__":
	"""
	This code block will only be executed if gmapstats.py is called as a script
	(ie. it will not be executed if the gmapstats class is just imported by another module)
	"""
	# build the script object
	engine=gmapstats()

	# run execute the main method
	engine.main()
